from helferklassen import Knoten, KantenSymbol, Lesen
from graphics_and_games_klassen import Ereignisbehandlung
import time

class GraphMatrix(Ereignisbehandlung):

    def __init__(self):
        """
        Baut die Datenstruktur auf
        """
        super().__init__()
        self.knoten = []
        self.kanten = []
        self.matrix = []
        self.besuchteKnoten = []
        self.weiter = False
        self.aktuellerWeg = []
        self.optimalerWeg = []
        
    def TasteGedrueckt(self, taste):
        """
        Die Aktionsmethode für gedrückte Tasten.
        :param taste: die gedrückte Taste
        """
        self.weiter = True

    def AufTasteWarten(self):
        """
        Wartet auf einen Tastendruck
        """
        while not self.weiter:
            time.sleep(0.1)
        self.weiter = False
        

    def KnotenEinfuegen(self, bezeichner, x, y):
        """
        Einfügen eines neuen Knoten in den Graphen
        """
        self.knoten.append(Knoten(bezeichner, x, y))
        for zeile in self.matrix:
            zeile.append(-1)
        self.matrix.append([])
        for zeile in self.matrix:
            self.matrix[-1].append(-1)

    def KnotenGeben(self, bezeichner):
        """
        Gibt den Knoten mit dem angegebenen Bezeichner zurück.
        """
        for k in self.knoten:
            if k.BezeichnerGeben() == bezeichner:
                return k
        return None
 
    def KnotenNummerGeben(self, bezeichner):
        """
        Gibt die interne Nummer des Knotens
        """
        for i in range(len(self.knoten)):
            if self.knoten[i].BezeichnerGeben() == bezeichner:
                return i
        return -1

    def KnotenBezeichnerGeben(self, KnotenNummer):
        """
        Gibt die Bezeichnung eines Knotens mit der internen Knotennummer
        """
        if knotenNummer < len(self.knoten) and knotenNummer >= 0:
            return self.knoten[knotenNummer].BezeichnerGeben()
        else:
            return ""

    def KanteEinfuegen(self, von, nach, gewichtung):
        """
        Einfügen einer Kante in den Graphen
        """
        vonNummer = self.KnotenNummerGeben(von)
        nachNummer = self.KnotenNummerGeben(nach)
        if vonNummer != -1 and nachNummer != -1 and vonNummer != nachNummer:
            self.matrix[vonNummer][nachNummer] = gewichtung
            self.matrix[nachNummer][vonNummer] = gewichtung
            self.kanten.append(KantenSymbol(self.knoten[vonNummer].SymbolGeben(), self.knoten[nachNummer].SymbolGeben(), False, str(gewichtung), 3, "blau"))

    def Ausgeben(self):
        """
        Gibt die Adjazenzmatrix des Graphen in der Konsole aus
        """
        # Kopfzeile
        print("    ", end='')
        for k in self.knoten:
            print((k.BezeichnerGeben() + "    ")[0:4], end='')
        print("")

        for i in range(len(self.knoten)):
            print((self.knoten[i].BezeichnerGeben() + "    ")[0:4], end='')
            for j in range(len(self.knoten)):
                if self.matrix[i][j] != -1:
                    print((str(self.matrix[i][j]) + "   ")[0:4], end='')
                else:
                    print("    ", end='')
            print("")

    def KnotenAnzahlGeben(self):
        """
        Gibt die Anzahl der Knoten des Graphen zurück
        """
        return len(self.knoten)

    def KanteGewichtGeben(self, von, nach):
        """
        Gibt die Gewichtung einer Kante zurück
        Die Kante ist durch einen Anfangsknoten und einen Endknoten festgelegt
        """
        vonNummer = self.KnotenNummerGeben(von)
        nachNummer = self.KnotenNummerGeben(nach)
        if vonNummer != -1 and nachNummer!=-1:
            return self.matrix[vonNummer][nachNummer]
        else:
            return -1

    def KantenfarbeSetzen(self, k1, k2, farbe):
        """
        Setzt die Farbe der Kante von Knoten k1 zu Knoten k2.
        :param k1: der eine Endknoten
        :param k2: der andere Endknoten
        :param farbe: die neue Kantenfarbe
        """
        for k in self.kanten:
            if ((k1.SymbolGeben() == k.StartsymbolGeben() and k2.SymbolGeben() == k.ZielsymbolGeben()) or
                (k2.SymbolGeben() == k.StartsymbolGeben() and k1.SymbolGeben() == k.ZielsymbolGeben())):
                k.FarbeSetzen(farbe)

    def ZurueckSetzen(self):
        """
        Löscht die Kanten und Knoten des Graphen.
        Die Anzeige wird auch gelöscht.
        """
        self.besuchteKnoten.clear()
        self.aktuellerWeg.clear()
        self.optimalerWeg.clear()
        self.optimaleLänge = float('inf')

    def OptimalerWegAnzeigen(self):
        """
        Zeigt den optimalen Weg an.
        """
        letzter = None
        for aktuell in self.optimalerWeg:
            aktuell.FarbeSetzen("rot")
            if letzter is not None:
                self.KantenfarbeSetzen(aktuell, letzter, "rot")
            letzter = aktuell

    def KnotenBesuchen1(self, aktuell, ziel):
        """
        Setzt den Rekursionsschritt bei einem Knoten um.
        Alle Wege werden gesucht.
        """
        self.besuchteKnoten.append(self.knoten[aktuell])
        self.knoten[aktuell].FarbeSetzen("gruen")
        if aktuell == ziel:
            self.AufTasteWarten()
        else:
            for nummer in range(len(self.knoten)):
                if (self.matrix[aktuell][nummer] > 0 and
                        self.knoten[nummer] not in self.besuchteKnoten):
                    self.KantenfarbeSetzen(self.knoten[aktuell], self.knoten[nummer], "gruen")
                    self.KnotenBesuchen1(nummer, ziel)
                    self.KantenfarbeSetzen(self.knoten[aktuell], self.knoten[nummer], "blau")
        self.knoten[aktuell].FarbeSetzen("weiss")
        self.besuchteKnoten.remove(self.knoten[aktuell])

    def KnotenBesuchen2(self, aktuell, ziel, länge):
        """
        Setzt den Rekursionsschritt bei einem Knoten um.
        Der optimale Weg wird gespeichert.
        """
        self.besuchteKnoten.append(self.knoten[aktuell])
        self.knoten[aktuell].FarbeSetzen("gruen")
        self.aktuellerWeg.append(self.knoten[aktuell])
        if aktuell == ziel:
            self.AufTasteWarten()
            if länge < self.optimaleLänge:
                self.optimalerWeg = list(self.aktuellerWeg)
                self.optimaleLänge = länge
        else:
            for nummer in range(len(self.knoten)):
                if (self.matrix[aktuell][nummer] > 0 and
                        self.knoten[nummer] not in self.besuchteKnoten):
                    self.KantenfarbeSetzen(self.knoten[aktuell], self.knoten[nummer], "gruen")
                    self.KnotenBesuchen2(nummer, ziel, länge + self.matrix[aktuell][nummer])
                    self.KantenfarbeSetzen(self.knoten[aktuell], self.knoten[nummer], "blau")
        self.knoten[aktuell].FarbeSetzen("weiss")
        self.besuchteKnoten.remove(self.knoten[aktuell])
        self.aktuellerWeg.pop()

class Wege:
    def __init__(self):
        """
        Legt das Graphenobjekt an.
        """
        self.g = GraphMatrix()
        self.l = Lesen()
        self.l.LesenDatenbank("Autobahn.grdb", self.g)

    def ReiseStarten1(self, start, ziel):
        """
        Startet die Besuche der Knoten.
        """
        self.g.ZurueckSetzen()
        self.g.KnotenBesuchen1(self.g.KnotenNummerGeben(start), self.g.KnotenNummerGeben(ziel))

    def ReiseStarten2(self, start, ziel):
        """
        Startet die Besuche der Knoten.
        """
        self.g.ZurueckSetzen()
        self.g.KnotenBesuchen2(self.g.KnotenNummerGeben(start), self.g.KnotenNummerGeben(ziel), 0)
        self.g.OptimalerWegAnzeigen()

#Test
w = Wege()
w.ReiseStarten1("M","N")
#w.ReiseStarten2("M","N")