from helferklassen import Knoten, KantenSymbol, Lesen
from graphics_and_games_klassen import Ereignisbehandlung
import time

class GraphMatrix(Ereignisbehandlung):

    def __init__(self):
        """
        Baut die Datenstruktur auf
        """
        super().__init__()
        self.knoten = []
        self.kanten = []
        self.matrix = []
        self.besuchteKnoten = []
        self.weiter = True

    def KnotenEinfuegen(self, bezeichner, x, y):
        """
        Einfügen eines neuen Knoten in den Graphen
        """
        self.knoten.append(Knoten(bezeichner, x, y))
        for zeile in self.matrix:
            zeile.append(-1)
        self.matrix.append([])
        for zeile in self.matrix:
            self.matrix[-1].append(-1)

    def KnotenGeben(self, bezeichner):
        """
        Gibt den Knoten mit dem angegebenen Bezeichner Zurueck.
        """
        for k in self.knoten:
            if k.BezeichnerGeben() == bezeichner:
                return k
        return None
 
    def KnotenNummerGeben(self, bezeichner):
        """
        Gibt die interne Nummer des Knotens
        """
        for i in range(len(self.knoten)):
            if self.knoten[i].BezeichnerGeben() == bezeichner:
                return i
        return -1

    def KnotenBezeichnerGeben(self, KnotenNummer):
        """
        Gibt die Bezeichnung eines Knotens mit der internen Knotennummer
        """
        if knotenNummer < len(self.knoten) and knotenNummer >= 0:
            return self.knoten[knotenNummer].BezeichnerGeben()
        else:
            return ""

    def KanteEinfuegen(self, von, nach, gewichtung):
        """
        Einfügen einer Kante in den Graphen
        """
        vonNummer = self.KnotenNummerGeben(von)
        nachNummer = self.KnotenNummerGeben(nach)
        if vonNummer != -1 and nachNummer != -1 and vonNummer != nachNummer:
            self.matrix[vonNummer][nachNummer] = gewichtung
            self.matrix[nachNummer][vonNummer] = gewichtung
            self.kanten.append(KantenSymbol(self.knoten[vonNummer].SymbolGeben(), self.knoten[nachNummer].SymbolGeben(), False, str(gewichtung), 3, "blau"))

    def Ausgeben(self):
        """
        Gibt die Adjazenzmatrix des Graphen in der Konsole aus
        """
        # Kopfzeile
        print("    ", end='')
        for k in self.knoten:
            print((k.BezeichnerGeben() + "    ")[0:4], end='')
        print("")

        for i in range(len(self.knoten)):
            print((self.knoten[i].BezeichnerGeben() + "    ")[0:4], end='')
            for j in range(len(self.knoten)):
                if self.matrix[i][j] != -1:
                    print((str(self.matrix[i][j]) + "   ")[0:4], end='')
                else:
                    print("    ", end='')
            print("")

    def KnotenAnzahlGeben(self):
        """
        Gibt die Anzahl der Knoten des Graphen Zurueck
        """
        return len(self.knoten)

    def KanteGewichtGeben(self, von, nach):
        """
        Gibt die Gewichtung einer Kante Zurueck
        Die Kante ist durch einen Anfangsknoten und einen Endknoten festgelegt
        """
        vonNummer = self.KnotenNummerGeben(von)
        nachNummer = self.KnotenNummerGeben(nach)
        if vonNummer != -1 and nachNummer!=-1:
            return self.matrix[vonNummer][nachNummer]
        else:
            return -1


    def ZurueckSetzen(self):
        """Löscht die Kanten und Knoten des Graphen. Die Anzeige wird auch gelöscht."""
        for k in self.kanten:
            k.Entfernen()
        for k in self.knoten:
            k.SymbolGeben().Entfernen()
        self.knoten.clear()
        self.kanten.clear()
        self.matrix.clear()

    def KnotenBesuchen(self, aktuell):
        """Setzt den Rekursionsschritt bei einem Knoten um."""
        print(f"Knoten: {self.knoten[aktuell].BezeichnerGeben()}")
        self.besuchteKnoten.append(aktuell)
        for nummer in range(len(self.knoten)):
            if self.matrix[aktuell][nummer] > 0 and nummer not in self.besuchteKnoten:
                self.KnotenBesuchen(nummer)
                print(f"Zurück nach: {self.knoten[aktuell].BezeichnerGeben()}")
        self.knoten[aktuell].FarbeSetzen("grün")

    def TiefensucheAusführen(self, startKnoten):
        """Führt die Tiefensuche vom Startknoten aus durch."""
        self.besuchteKnoten.clear()
        self.KnotenBesuchen(startKnoten)

    def ZusammenhangTesten(self):
        """Testet, ob der Graph zusammenhängend ist."""
        self.besuchteKnoten.clear()
        self.KnotenBesuchen(0)
        return len(self.besuchteKnoten) == len(self.knoten)


class Beispiele:
    """Beispiele für diverse Graphen."""

    def __init__(self):
        """Legt das Graphenobjekt an."""
        self.g = GraphMatrix()
        self.l = Lesen()

    def AusfuehrenAbiturfahrt(self):
        """Legt einen Graphen für eine Abiturfahrt an."""
        self.g.ZurueckSetzen()
        if self.l.LesenDatenbank("Abiturfahrt.grdb", self.g):
            pass  # g.Ausgeben()

    def AusfuehrenAutobahn(self):
        """Legt einen Graphen für den süddeutschen Teil des Autobahnnetzes an."""
        self.g.ZurueckSetzen()
        if self.l.LesenDatenbank("Autobahn.grdb", self.g):
            pass  # g.Ausgeben()

    def AusfuehrenICE(self):
        """Legt einen Graphen für das deutsche ICE-Netz an."""
        self.g.ZurueckSetzen()
        if self.l.LesenDatenbank("ICENetz.grdb", self.g):
            pass  # g.Ausgeben()

    def AusfuehrenBeispielgraph(self):
        """Legt einen zusammenhängenden Graphen an."""
        self.g.ZurueckSetzen()
        if self.l.LesenDatenbank("Beispielgraph.grdb", self.g):
            pass  # g.Ausgeben()

    def AusfuehrenBeispielgraph2(self):
        """Legt einen nicht zusammenhängenden Graphen an."""
        self.g.ZurueckSetzen()
        if self.l.LesenDatenbank("Beispielgraph2.grdb", self.g):
            pass  # g.Ausgeben()

    def TiefensucheAusfuehren(self, nameStartKnoten):
        """Führt die Tiefensuche vom Startknoten aus durch."""
        if self.g:
            self.g.TiefensucheAusführen(self.g.KnotenNummerGeben(nameStartKnoten))

    def ZusammenhangTesten(self):
        """Testet, ob der Graph zusammenhängend ist."""
        if self.g.ZusammenhangTesten():
            print("Der Graph ist zusammenhängend.")
        else:
            print("Der Graph ist nicht zusammenhängend.")

# Anwendung
b = Beispiele()
b.AusfuehrenBeispielgraph2()
b.TiefensucheAusfuehren("A")